#include <sys/types.h>
#include <sys/cpuset.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <pthread.h>
#include <pthread_np.h>
#include <unistd.h>
#include <bsock/bsock.h>

#include "logger.h"
#include "mod.h"
#include "msg.hh"
#include "util.h"
#include "io.hh"

#include <cerrno>
#include <csignal>
#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <sstream>
#include <unordered_map>
#include <vector>

LOGGER_VAR_DECL;

static constexpr int NEVENT = 64;
static constexpr int SOCK_BACKLOG = 10000;
static constexpr int SINGLE_LEGACY = -1;
static constexpr int DEFAULT_PORT = 9898;
static constexpr int BSOCK_BUF_SZ = 4096;
// 16MB max per message
static constexpr int MAX_MODE_PARAMS = 16;
static constexpr int MAX_MODE_PARAMS_LEN = 128;

struct ppd_options {
	int port;
	int shared_kq;
	cpuset_t cpuset;
	int verbose;
	int enable_tls;
	int enable_ktls;
	char secret_key_fn[128];
	char cert_fn[128];

	std::vector<char *> hpip;

	int skq_dump;
	int skq_flag;
	int skq_rtshare;
	int skq_rtfreq;

	/* the mode this server runs in */
	char module_path[128];

	char *mod_argk[MAX_MODE_PARAMS];
	char *mod_argv[MAX_MODE_PARAMS];
	int mod_argc;

	/* runtime states */
	SSL_CTX *ssl_ctx; // static SSL_CTX
	struct ppd_mod_info *m_info;
	void *m_global_ctx;
};

struct ppd_conn {
	struct bsock * bsock;
	int conn_fd;

	SSL * ssl;
	char * ssl_readbuf;
	struct ppd_bsock_io_ssl_ctx ssl_io_ctx;
	void *m_conn_ctx;
};

struct ppd_thread_ctx {
	int tid;
	pthread_t thrd;
	int kqfd;
	long evcnt;
	void *m_thread_ctx;
	char *m_buf;
};

static ppd_options options = { .port = DEFAULT_PORT,
	.shared_kq = 0,
	.cpuset = CPUSET_T_INITIALIZER(1), // first core
	.verbose = 0,
	.enable_tls = 0,
	.enable_ktls = 0,
	.secret_key_fn = { 0 },
	.cert_fn = { 0 },
	.skq_dump = 0,
	.skq_flag = 0,
	.skq_rtshare = 100,
	.skq_rtfreq = 0,
	.module_path = { 0 },
	.mod_argc = 0,
	.ssl_ctx = nullptr,
	.m_info = nullptr };

static int
main_kqueue_create(std::vector<int> *server_socks)
{
	int status;
	struct kevent kev;

	int mkqfd = kqueue();
	if (mkqfd == -1) {
		E("Failed to create kqueue: %d.\n", errno);
	}

	for (uint32_t i = 0; i < server_socks->size(); i++) {
		EV_SET(&kev, server_socks->at(i), EVFILT_READ, EV_ADD, 0, 0, NULL);

		status = kevent(mkqfd, &kev, 1, NULL, 0, NULL);
		if (status == -1) {
			E("kevent() failed: %d.\n", errno);
		}
	}

#define TIMER_FD (-1234)
	EV_SET(&kev, TIMER_FD, EVFILT_TIMER, EV_ADD, NOTE_MSECONDS, 1, NULL);
	status = kevent(mkqfd, &kev, 1, NULL, 0, NULL);
	if (status == -1) {
		E("kevent() timer failed: %d.\n", errno);
	}

	return mkqfd;
}

static void
create_tls_context()
{
	SSL_CTX *ctx;

	ctx = SSL_CTX_new(TLS_server_method());
	if (!ctx) {
		E("SSL_CTX_new() failed: %ld\n", ERR_get_error());
	}

	SSL_CTX_set_min_proto_version(ctx, TLS1_2_VERSION);
	SSL_CTX_set_max_proto_version(ctx, TLS1_2_VERSION);

	if (options.enable_ktls) {
		if ((SSL_CTX_set_options(ctx, SSL_OP_ENABLE_KTLS) & SSL_OP_ENABLE_KTLS) !=
		    SSL_OP_ENABLE_KTLS) {
			E("SSL_CTX_set_options() SSL_OP_ENABLE_KTLS failed.\n");
		}
	}

	/* Set the key and cert */
	if (SSL_CTX_use_certificate_file(ctx, options.cert_fn, SSL_FILETYPE_PEM) <= 0) {
		E("SSL_CTX_use_certificate_file() %s failed: %ld\n", options.cert_fn,
		    ERR_get_error());
	}

	if (SSL_CTX_use_PrivateKey_file(ctx, options.secret_key_fn, SSL_FILETYPE_PEM) <= 0) {
		E("SSL_CTX_use_PrivateKey_file() %s failed: %ld\n", options.secret_key_fn,
		    ERR_get_error());
	}

	options.ssl_ctx = ctx;
}

static void
listen_socket_create(std::vector<int> *socks)
{
	struct sockaddr_in server_addr;
	int status;
	const int enable = 1;
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(options.port);
	server_addr.sin_addr.s_addr = INADDR_ANY;

	for (int i = 0; i < 1; i++) {
		int fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (fd < 0) {
			E("socket() returned %d\n", errno);
		}

		if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
			E("setsockopt() reuseaddr %d\n", errno);
		}

		if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &enable, sizeof(enable)) < 0) {
			E("setsockopt() reuseport %d\n", errno);
		}

		if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &enable, sizeof(enable)) < 0) {
			E("setsockopt() NODELAY %d\n", errno);
		}

		status = bind(fd, (struct sockaddr *)&server_addr, sizeof(server_addr));
		if (status < 0) {
			E("bind() returned %d\n", errno);
		}

		status = listen(fd, SOCK_BACKLOG);
		if (status < 0) {
			E("listen() returned %d\n", errno);
		}

		socks->push_back(fd);
	}
}

static void
ppd_conn_free_no_ctx(struct ppd_conn *conn)
{
	if (conn->ssl != nullptr) {
		SSL_shutdown(conn->ssl);
		SSL_free(conn->ssl);
	}
	if (conn->bsock != nullptr) {
		bsock_free(conn->bsock);
	}
	if (conn->ssl_readbuf != nullptr) {
		delete[] conn->ssl_readbuf;
	}
	close(conn->conn_fd);
	delete conn;
}

static void
ppd_conn_free(struct ppd_thread_ctx *tinfo, struct ppd_conn *conn)
{
	void *conn_ctx = conn->m_conn_ctx;
	ppd_conn_free_no_ctx(conn);
	options.m_info->conn_free_cb(options.m_global_ctx, tinfo->m_thread_ctx, conn_ctx);
}

static void
drop_conn(struct ppd_thread_ctx *tinfo, struct kevent *kev)
{
	int status;
	int conn_fd = kev->ident;
	struct kevent ev;
	EV_SET(&ev, conn_fd, EVFILT_READ, EV_DELETE, 0, 0, NULL);

	status = kevent(tinfo->kqfd, &ev, 1, 0, 0, NULL);

	if (status < 0) {
		E("Failed to drop connection %d from kqueue: %d\n", conn_fd, errno);
	}

	ppd_conn_free(tinfo, (struct ppd_conn *)kev->udata);
}

static int
handle_event(struct ppd_thread_ctx *tinfo, struct kevent *kev)
{
	int conn_fd;
	int status;
	struct ppd_conn *hint;

	if (kev->filter != EVFILT_READ) {
		E("Unknown event filter %d\n", kev->filter);
	}

	conn_fd = kev->ident;
	hint = (struct ppd_conn *)kev->udata;
	struct ppd_msg *msg;
	size_t out_sz;

	if (kev->flags & EV_EOF) {
		V("Thread %d dropped connection %d due to EOF.\n", tinfo->tid, conn_fd);
		status = ECONNRESET;
		goto fail;
	}

	// read data first 
	status = bsock_poll(hint->bsock);
	if (status == 0) {
		// connection reset basically
		W("Thread %d dropped connection %d due to bsock_poll ret %d errno %d\n", tinfo->tid, conn_fd, status, errno);
		goto fail;
	}

	while (true) {
		status = ppd_readmsg(hint->bsock, tinfo->m_buf, PPD_MSG_MAX_SZ);
		if (status != 0) {
			if (errno == ENODATA) {
				// not enough data yet. try again later.
				goto end;
			} else {
				W("Thread %d dropped connection %d due to ppd_readmsg error %d\n", tinfo->tid, conn_fd, errno);
				goto fail;
			}
		}

		msg = (struct ppd_msg *)tinfo->m_buf;
		D("Thread %d connection %d RECV msg of size %u\n", tinfo->tid, conn_fd, msg->size);
		status = options.m_info->conn_recv_cb(msg->data, msg->size, options.m_global_ctx,
			tinfo->m_thread_ctx, hint->m_conn_ctx);
		if (status != 0) {
			W("Thread %d dropped connection %d due to conn_recv_cb error %d\n", tinfo->tid,
				conn_fd, status);
			goto fail;
		}

		status = options.m_info->conn_send_cb(msg->data, PPD_MSG_MAX_SZ - PPD_MSG_HDR_SZ, &out_sz,
			options.m_global_ctx, tinfo->m_thread_ctx, hint->m_conn_ctx);
		if (status != 0) {
			W("Thread %d dropped connection %d due to conn_send_cb error %d\n", tinfo->tid,
				conn_fd, status);
			goto fail;
		}

		msg->size = out_sz;
		status = ppd_writemsg(hint->bsock, msg);
		if (status != 0) {
			// shouldn't be error here unless msg is too big to fit in bsock buffer
			W("Thread %d dropped connection %d due to ppd_writemsg error %d\n", tinfo->tid, conn_fd, errno);
			goto fail;
		}

		D("Thread %d connection %d SEND msg of size %lu\n", tinfo->tid, conn_fd, out_sz);
		// flush bsock immediately
		status = bsock_flush(hint->bsock);
		if (status < 0) {
			W("Thread %d dropped connection %d due to bsock_flush ret %d errno %d\n", tinfo->tid, conn_fd, status, errno);
			goto fail;
		}
	}

	tinfo->evcnt++;
end:
	return 0;
fail:
	drop_conn(tinfo, kev);
	return status;
}

static void *
worker_main(void *info)
{
	struct ppd_thread_ctx *tinfo = (struct ppd_thread_ctx *)info;
	struct kevent kev[NEVENT];
	struct kevent skev[NEVENT];
	int skev_sz;
	int status;

	V("Thread %d started.\n", tinfo->tid);

	skev_sz = 0;
	while (true) {
		status = kevent(tinfo->kqfd, skev, skev_sz, kev, NEVENT, NULL);

		if (status <= 0) {
			E("kevent() failed: %d\n", errno);
		}

		skev_sz = 0;
		for (int i = 0; i < status; i++) {
			if (handle_event(tinfo, &kev[i]) == 0) {
				if (options.shared_kq && options.skq_flag == SINGLE_LEGACY) {
					EV_SET(&skev[skev_sz], (int)kev[i].ident, EVFILT_READ,
					    EV_ENABLE, 0, 0, kev[i].udata);
					skev_sz++;
				}
			}
		}
	}
}

static void
dump_options()
{
	std::stringstream ss;
	ss << "Configuration:\n"
	   << "        port: " << options.port << std::endl
	   << "        module: " << options.module_path << std::endl
	   << "        enable TLS: " << options.enable_tls << std::endl
	   << "        enable kTLS: " << options.enable_ktls << std::endl
	   << "        TLS secret key: " << options.secret_key_fn << std::endl
	   << "        TLS certificate: " << options.cert_fn << std::endl
	   << "        shared Kqueue: " << options.shared_kq << std::endl
	   << "        number of threads: " << CPU_COUNT(&options.cpuset) << std::endl
	   << "        verbose: " << options.verbose << std::endl
	   << "        SKQ flags: " << options.skq_flag << std::endl
	   << "        SKQ dump:" << options.skq_dump << std::endl
	   << "        SKQ rtshare: " << options.skq_rtshare << std::endl
	   << "        SKQ rtfreq: " << options.skq_rtfreq << std::endl
	   << "        SKQ high priority clients (" << options.hpip.size() << "): " << std::endl;

	for (uint i = 0; i < options.hpip.size(); i++) {
		ss << "                " << options.hpip.at(i) << std::endl;
	}

	V("%s", ss.str().c_str());
}

static void
usage()
{
	fprintf(stdout,
	    "Usage:\n"
	    "    p: server port\n"
	    "    v: verbose mode\n"
	    "    m: shared Kqueue across workers\n"
	    "    s: enable TLS\n"
	    "    k: enable kTLS\n"
	    "    S: TLS private key file\n"
	    "    C: TLS certificate file\n"
	    "    c: server thread cpu list\n"
	    "    h: show help\n"
	    "    d: SKQ dump interval (s)\n"
	    "    r: SKQ realtime client hostname(s)\n"
	    "    R: SKQ rt_share\n"
	    "    F: SKQ rt_freq\n"
	    "    M: module path\n"
	    "    O: module specific parameters of format \"key=value\"\n");
}

static void
create_workers(std::vector<struct ppd_thread_ctx *> *workers)
{
	int kq = -1;

	if (options.shared_kq) {
		kq = kqueue();

		if (kq <= 0) {
			E("Failed to create kqueue. ERR %d\n", errno);
		}

#ifdef FKQMULTI
		if (options.skq_flag != SINGLE_LEGACY) {
			status = ioctl(kq, FKQMULTI, &options.skq_flag);
			if (status == -1) {
				E("ioctl failed. ERR %d\n", errno);
			}
			V("SKQ enabled. SFLAG %d\n", options.skq_flag);
		}

		int para = KQTUNE_MAKE(KQTUNE_RTSHARE, options.kq_rtshare);
		status = ioctl(kq, FKQTUNE, &para);
		if (status == -1) {
			E("rtshare ioctl failed. ERR %d\n", errno);
		}

		para = KQTUNE_MAKE(KQTUNE_FREQ, options.kq_tfreq);
		status = ioctl(kq, FKQTUNE, &para);
		if (status == -1) {
			E("freq ioctl failed. ERR %d\n", errno);
		}
		V("KQ IOCTL: RTSHARE %d TFREQ %d\n", options.kq_rtshare, options.kq_tfreq);
#else
		/* legacy single KQ only supports flag -1 */
		options.skq_flag = SINGLE_LEGACY;
#endif
	}

	int tid = 0;
	int core;
	CPU_FOREACH_ISSET (core, &options.cpuset) {
		V("Creating worker thread on core %d...\n", core);

		struct ppd_thread_ctx *thrd = (struct ppd_thread_ctx *)
		    aligned_alloc(CACHE_LINE_SIZE, sizeof(struct ppd_thread_ctx));
		thrd->evcnt = 0;
		thrd->tid = tid;
		options.m_info->thread_init_cb(core, options.m_global_ctx, &thrd->m_thread_ctx);

		if (!options.shared_kq) {
			kq = kqueue();
			if (kq <= 0) {
				E("kqueue() failed: %d\n", errno);
			}

#ifdef FKQMULTI
			int para = KQTUNE_MAKE(KQTUNE_RTSHARE, options.kq_rtshare);
			status = ioctl(kq, FKQTUNE, &para);
			if (status == -1) {
				E("rtshare ioctl failed. ERR %d\n", errno);
			}

			para = KQTUNE_MAKE(KQTUNE_FREQ, options.kq_tfreq);
			status = ioctl(kq, FKQTUNE, &para);
			if (status == -1) {
				E("freq ioctl failed. ERR %d\n", errno);
			}
#endif
		}

		thrd->kqfd = kq;
		thrd->m_buf = new char[PPD_MSG_MAX_SZ];

		pthread_attr_t attr;
		int status;
		if ((status = pthread_attr_init(&attr)) != 0) {
			E("pthread_attr_init() failed: %d\n", status);
		}

		cpuset_t cpuset;
		CPU_ZERO(&cpuset);
		CPU_SET(core, &cpuset);

		status = pthread_attr_setaffinity_np(&attr, sizeof(cpuset), &cpuset);
		if (status != 0) {
			E("pthread_attr_setaffinity_np() failed : %d\n", status);
		}

		status = pthread_create(&thrd->thrd, &attr, worker_main, thrd);
		if (status != 0) {
			E("pthread_create() failed: %d\n", status);
		}
		workers->push_back(thrd);

		tid++;
	}
}

static void
parse_mod_arg(char *arg, char *argk, size_t szk, char *argv, size_t szv)
{
	char *saveptr = NULL;

	char *key = strtok_r(arg, "=", &saveptr);
	char *val = strtok_r(NULL, "=", &saveptr);

	V("Parsed module arg: %s = %s\n", key, val);
	snprintf(argk, szk, "%s", key);
	snprintf(argv, szv, "%s", val);
}

static SSL *
tls_handshake_server(int conn_fd)
{
	SSL *ssl;
	int r;

	ssl = SSL_new(options.ssl_ctx);
	if (ssl == NULL) {
		E("SSL_new() failed: %ld\n", ERR_get_error());
	}

	if (SSL_set_fd(ssl, conn_fd) == 0) {
		E("SSL_set_fd() failed: %ld\n", ERR_get_error());
	}

	if ((r = SSL_accept(ssl)) <= 0) {
		E("SSL_accept() failed: %ld\n", ERR_get_error());
	}

	if (options.enable_ktls) {
		// make sure ktls is enabled
		if (!(BIO_get_ktls_send(SSL_get_wbio(ssl)) &&
			BIO_get_ktls_recv(SSL_get_rbio(ssl)))) {
			E("kTLS not enabled on %d: %ld\n", conn_fd, ERR_get_error());
		}
	}

	return ssl;
}

static void
loop_main(int m_kq, std::vector<struct ppd_thread_ctx *> *workers)
{
	struct kevent kev;
	int cur_conn = 0;
#ifdef FKQMULTI
	int cur_ts = 0;
#endif
	while (1) {
		if (kevent(m_kq, NULL, 0, &kev, 1, NULL) != 1) {
			E("kevent() failed: %d\n", errno);
		}

		if (kev.ident == (uintptr_t)TIMER_FD) {
#ifdef FKQMULTI
			/* timer event */
			if (options.skq_dump > 0 && cur_ts % (options.skq_dump * 1000) == 0) {
				uintptr_t args = (uintptr_t)dbuf;
				memset(dbuf, 0, 1024 * 1024 + 1);
				status = ioctl(kq, FKQMPRNT, &args);
				if (status == -1) {
					E("SKQ dump failed %d\n", errno);
				} else {
					fprintf(stdout, "====== KQ DUMP ======\n%s\n", dbuf);
				}
			}
			cur_ts++;
#endif
			continue;
		}

		// new connection
		struct sockaddr client_addr;
		socklen_t client_addr_size = sizeof(client_addr);

		int conn_fd = accept(kev.ident, &client_addr, &client_addr_size);

		if (conn_fd < 0) {
			W("accept() failed: %d\n", errno);
			continue;
		}

		{
			const int enable = 1;
			if (setsockopt(conn_fd, IPPROTO_TCP, TCP_NODELAY, &enable, sizeof(enable)) <
			    0) {
				W("setsockopt() nodelay failed on conn %d: err %d\n", conn_fd,
				    errno);
				close(conn_fd);
				continue;
			}
		}

		if (getpeername(conn_fd, &client_addr, &client_addr_size) < 0) {
			E("getpeername failed: %d\n", errno);
		}

		char client_ip[INET_ADDRSTRLEN];
		if (inet_ntop(AF_INET, &(((struct sockaddr_in *)&client_addr)->sin_addr), client_ip, sizeof(client_ip)) == NULL) {
			E("inet_ntop failed: %d\n", errno);
		}

		V("Accepted connection %d from %s.\n", conn_fd, client_ip);

		struct ppd_conn *conn = new struct ppd_conn;

		conn->conn_fd = conn_fd;

		if (options.enable_tls) {
			conn->ssl = tls_handshake_server(conn_fd);
			conn->ssl_readbuf = new char[BSOCK_BUF_SZ];
			struct bsock_ringbuf_io io = ppd_bsock_io_ssl();
			conn->ssl_io_ctx.ssl_readbuf = conn->ssl_readbuf;
			conn->ssl_io_ctx.ssl = conn->ssl;
			conn->ssl_io_ctx.ssl_readbuf_len = BSOCK_BUF_SZ;
			V("Established TLS on connection %d...\n", conn_fd);
			conn->bsock = bsock_create((void*)&conn->ssl_io_ctx, &io, BSOCK_BUF_SZ, 0);
		} else {
			conn->ssl = nullptr;
			conn->ssl_readbuf = nullptr;
			struct bsock_ringbuf_io io = bsock_io_posix();
			conn->bsock = bsock_create((void*)(uintptr_t)conn_fd, &io, BSOCK_BUF_SZ, 0);
		}

		if (conn->bsock == nullptr) {
			W("Failed to create bsock on connection %d...\n", conn_fd);
			ppd_conn_free_no_ctx(conn);
			continue;
		}

		int worker_idx = cur_conn % workers->size();
		if (options.m_info->conn_init_cb(options.m_global_ctx,
			workers->at(worker_idx)->m_thread_ctx, &conn->m_conn_ctx) != 0) {
			W("Failed to create connection %d ctx, dropped.\n", conn_fd);
			ppd_conn_free_no_ctx(conn);
			continue;
		}

		int ev_flags = EV_ADD;
#ifdef FKQMULTI
		for (uint32_t i = 0; i < options.hpip.size(); i++) {
			if (strcmp(ipaddr, options.hpip.at(i)) == 0) {
				V("Connection %d marked as realtime.\n", conn_fd);
				ev_flags |= EV_REALTIME;
			}
		}
#endif

		if (options.shared_kq && options.skq_flag == SINGLE_LEGACY) {
			ev_flags |= EV_DISPATCH;
		}

		int target_kq = workers->at(worker_idx)->kqfd;
		EV_SET(&kev, conn_fd, EVFILT_READ, ev_flags, 0, 0, conn);
		if (kevent(target_kq, &kev, 1, NULL, 0, NULL) == -1) {
			E("kevent() failed: %d\n", errno);
		}

		V("Connection %d assigned to thread %d\n", conn_fd,
		    workers->at(cur_conn % workers->size())->tid);
		cur_conn++;
	}
}

int
main(int argc, char *argv[])
{
	SSL_library_init();
	srand(time(NULL));
	// don't raise SIGPIPE when sending into broken TCP connections
	::signal(SIGPIPE, SIG_IGN);

#ifdef FKQMULTI
	char dbuf[1024 * 1024 + 1];
#endif
	int ch;
	while ((ch = getopt(argc, argv, "d:p:c:m:vhr:R:F:M:O:skS:C:")) != -1) {
		switch (ch) {
		case 'd':
			options.skq_dump = atoi(optarg);
			break;
		case 'p':
			options.port = atoi(optarg);
			break;
		case 'c':
			cpulist_to_cpuset(optarg, &options.cpuset);
			break;
		case 'm':
			options.shared_kq = 1;
			options.skq_flag = atoi(optarg);
			break;
		case 'O': {
			if (options.mod_argc >= MAX_MODE_PARAMS) {
				E("Too many module args, max %d.\n", MAX_MODE_PARAMS);
			}
			options.mod_argk[options.mod_argc] = new char[MAX_MODE_PARAMS_LEN];
			options.mod_argv[options.mod_argc] = new char[MAX_MODE_PARAMS_LEN];
			parse_mod_arg(optarg, options.mod_argk[options.mod_argc],
			    MAX_MODE_PARAMS_LEN, options.mod_argv[options.mod_argc],
			    MAX_MODE_PARAMS_LEN);
			options.mod_argc++;
			break;
		}
		case 'M':
			snprintf(options.module_path, sizeof(options.module_path), "%s", optarg);
			break;
		case 'v':
			options.verbose++;
			if (options.verbose == INFO_VERBOSITY) {
				W("Verbose mode may cause SUBSTANTIAL latency fluctuation in some terminals emulators\n");
			} else if (options.verbose == DEBUG_VERBOSITY) {
				W("Debug mode enabled. Expect very verbose output.\n");
			}
			logger_set_verbosity(options.verbose);
			break;
		case 'r': {
			char *eip = new char[INET_ADDRSTRLEN + 1];
			if (nslookup(optarg, eip, INET_ADDRSTRLEN) != 0) {
				E("nslookup failed with %d for %s\n", errno, optarg);
			}
			options.hpip.push_back(eip);
			break;
		}
		case 'R':
			options.skq_rtshare = atoi(optarg);
			break;
		case 'F':
			options.skq_rtfreq = atoi(optarg);
			break;
		case 's':
			options.enable_tls = 1;
			break;
		case 'k':
			options.enable_ktls = 1;
			break;
		case 'S':
			snprintf(options.secret_key_fn, sizeof(options.secret_key_fn), "%s",
			    optarg);
			break;
		case 'C':
			snprintf(options.cert_fn, sizeof(options.cert_fn), "%s", optarg);
			break;
		case 'h':
			usage();
			exit(0);
		default:
			E("Unrecognized option -%c. See -h.\n\n", ch);
		}
	}
	dump_options();

	std::vector<struct ppd_thread_ctx *> wrk_thrds;
	std::vector<int> server_socks;

	if (strlen(options.module_path) == 0) {
		E("Must specify module path.\n");
	}

	V("Loading module %s...\n", options.module_path);
	options.m_info = ppd_load_module(options.module_path);
	V("Loaded module \"%s\".\n", options.m_info->name);

	V("Initializing module...\n");
	options.m_info->global_init_cb(options.mod_argc, options.mod_argk, options.mod_argv,
	    &options.m_global_ctx);

	V("Setting up server sockets...\n");
	listen_socket_create(&server_socks);

	if (options.enable_tls) {
		V("Setting up TLS context...\n");
		create_tls_context();
	}

	V("Setting up main thread kqueue...\n");
	int m_kq = main_kqueue_create(&server_socks);

	V("Creating worker threads...\n");
	create_workers(&wrk_thrds);

	V("Entering main event loop...\n");
	loop_main(m_kq, &wrk_thrds);

	E("Critical bug - main loop exited\n");
	return 0;
}
